; Macros for the bit-bang SPI for PIC18F processors
; Version 2.1  2008.06.04
; Inspired by the:
; http://www.piclist.org/techref/postbot.asp?by=time&id=piclist/2004/03/03/105839a
; This is the public domain code written by
; Wojciech M. Zabolotny wzab(at)ise.pw.edu.pl
; The code below requires you to define:
; sdo_port, sdo_pin: port and pin used as SDI
; sdi_port, sdi_pin: port and pin used as SDO
; clk_port, clk_pin: port and pin used as SCLK
; You have to set the proper pins appropriately for outputs or inputs
; If you want to use the fastest mspi_byte_wio routine, then you need
; to define a single byte spi_outdata variable in the access bank.
; All routines emulate the MODE0 (CPHA=0, CPOL=0) of the SPI 
; (as they are supposed to work with the SD/MMC memory card)
    list p=18f4550
#include <p18f4550.inc>
#include bit_bang_spi/mspi.inc
SPI_DATA UDATA_ACS
spi_outdata RES 1
    global spi_outdata
; Code section
MSPI_CODE CODE

;The MSPI_WIO_BIT macro outputs the bit from W7 to SPI,
;and reads the bit from SPI to W0,
;simultaneously the W is shifted by 1 bit to the left
MSPI_WIO_BIT macro
; move the highest bit to the carry, and 0 to the lowest bit
  addwf WREG,0,ACCESS
; set the data line to carry
  bsf   sdo_port,sdo_pin,ACCESS
  btfss STATUS,0,ACCESS
  bcf   sdo_port,sdo_pin,ACCESS
; read the input pin
  btfsc sdi_port, sdi_pin,ACCESS
; set the bit 0 in WREG to SDI
  bsf  WREG,0,ACCESS
; set the clock line to 'H'
  bsf  clk_port, clk_pin,ACCESS
; set the clock line to 'L'
  bcf  clk_port, clk_pin, ACCESS
  endm

;The MSPI_DIO_BIT macro outputs the particular bit from spi_outdata
;and reads the particular bit from SPI to WREG
;spi_outdata should be located in the access bank
MSPI_DIO_BIT macro bit_nr
; set the data line to the data bit
  bsf   sdo_port,sdo_pin,ACCESS
  btfss spi_outdata, bit_nr, ACCESS
  bcf   sdo_port,sdo_pin,ACCESS
; read the input pin
  btfsc sdi_port, sdi_pin,ACCESS
; set the bit 0 in WREG to SDI
  bsf  WREG,bit_nr,ACCESS
; set the clock line to 'H'
  bsf  clk_port, clk_pin,ACCESS
; set the clock line to 'L'
  bcf  clk_port, clk_pin, ACCESS
  endm

; Implementation changed, to allow keeping SDO high
; When sending 0xff
MSPI_O_BIT macro bit_nr
  bsf sdo_port, sdo_pin, ACCESS
  btfss WREG,bit_nr, ACCESS
  bcf sdo_port, sdo_pin, ACCESS
  nop
  bsf clk_port, clk_pin, ACCESS
  nop
  bcf clk_port, clk_pin, ACCESS
  endm

MSPI_I_BIT macro bit_nr
  btfsc sdi_port, sdi_pin, ACCESS
  bsf WREG, bit_nr, ACCESS
  bsf clk_port, clk_pin, ACCESS
  nop
  bcf clk_port, clk_pin, ACCESS
  nop
  endm

; Procedures
; mspi_byte_wio - sends to the SPI the byte from WREG
; reads the byte from SPI and puts it to WREG
mspi_byte_wio:
  global mspi_byte_wio
  MSPI_WIO_BIT ; 7
  MSPI_WIO_BIT ; 6
  MSPI_WIO_BIT ; 5
  MSPI_WIO_BIT ; 4
  MSPI_WIO_BIT ; 3
  MSPI_WIO_BIT ; 2
  MSPI_WIO_BIT ; 1
  MSPI_WIO_BIT ; 0
  return

; mspi_byte_dio - sends to the SPI the byte from spi_outdata
; reads the data from SPI and puts it to WREG
mspi_byte_dio:
  global mspi_byte_dio
  clrf WREG,0
  MSPI_DIO_BIT 7
  MSPI_DIO_BIT 6
  MSPI_DIO_BIT 5
  MSPI_DIO_BIT 4
  MSPI_DIO_BIT 3
  MSPI_DIO_BIT 2
  MSPI_DIO_BIT 1
  MSPI_DIO_BIT 0
  return

; mspi_byte_i - reads the byte from the SPI and puts it to the WREG
; the sdo pin remains unchanged (set it to the desired state before
; calling this procedure)
mspi_byte_i:
  global mspi_byte_i
  clrf WREG,0
  MSPI_I_BIT 7
  MSPI_I_BIT 6
  MSPI_I_BIT 5
  MSPI_I_BIT 4
  MSPI_I_BIT 3
  MSPI_I_BIT 2
  MSPI_I_BIT 1
  MSPI_I_BIT 0
  return

; mspi_byte_o - sends the byte from WREG to the SPI
mspi_byte_o:
  global mspi_byte_o
  MSPI_O_BIT 7
  MSPI_O_BIT 6
  MSPI_O_BIT 5
  MSPI_O_BIT 4
  MSPI_O_BIT 3
  MSPI_O_BIT 2
  MSPI_O_BIT 1
  MSPI_O_BIT 0
  return
end
