	list      p=18f4550            ; list directive to define processor
	#include <p18f4550.inc>        ; processor specific variable definitions
bank0 udata
buf_rd_tmp res 1
buf_wr_tmp res 1
buf_rd_ptr res 2
buf_wr_ptr res 2
	global buf_wr_ptr

buffer udata 0x300
buf_start res 0x100
BUF_END equ buf_start + 0xff
	extern sampling_flag

BUF_CODE CODE
; Subroutine for initialization of the buffer
buf_init:
	global buf_init
	banksel buf_rd_ptr
	movlw low(buf_start)
	movwf buf_rd_ptr,BANKED
	movlw high(buf_start)
	movwf buf_rd_ptr+1,BANKED
	banksel buf_wr_ptr
	movlw low(buf_start+1)
	movwf buf_wr_ptr,BANKED
	movlw high(buf_start+1)
	movwf buf_wr_ptr+1,BANKED
	return
; Subroutine for buffer writing from user context (not
; from interrupt)
buf_write_user:			
   global buf_write_user
	; block the interrupts
	bcf INTCON,GIE, ACCESS ; GIEH enabled
	nop  ; delay after clearing GIE (there was a bug in some PICs... 
	; next instruction could be interrupted)
	movff buf_wr_ptr,FSR2L
	movff buf_wr_ptr+1,FSR2H
   call  buf_write_always
	movff FSR2L,buf_wr_ptr
	movff FSR2H,buf_wr_ptr+1
	bsf INTCON,GIE, ACCESS ; GIEH enabled
   return

; Subroutine for buffer writing
buf_write:
	global buf_write
	; Check if the sampling_flag is on
	banksel sampling_flag
	btfss sampling_flag,0,BANKED
	retlw 0x00
buf_write_always: ; this entry point allows to write the data even
   ; when the sampling flag is cleared (essential for marking
   ; of start and end of recordings)
	movff WREG,buf_wr_tmp
	banksel buf_rd_ptr
	movf buf_rd_ptr,W,BANKED
	cpfseq FSR2L,ACCESS
	goto buf_write_ok
	movf buf_rd_ptr+1,W,BANKED
	cpfseq FSR2H,ACCESS
	goto buf_write_ok
	retlw 0xff
buf_write_ok:
	; Write the buffer
	movff buf_wr_tmp,POSTINC2
	; Now adjust the buffer
	movlw high(BUF_END+1)
	cpfseq FSR2H,ACCESS
	retlw 0x00
	movlw low(BUF_END+1)
	cpfseq FSR2L,ACCESS
	retlw 0x00
	movlw high(buf_start)
	movwf FSR2H,ACCESS
	movlw low(buf_start)
	movwf FSR2L,ACCESS
	retlw 0x00
; Subroutine for buffer reading
buf_read:
	global buf_read
	movff buf_rd_ptr,FSR2L
	movff buf_rd_ptr+1,FSR2H
	movff PREINC2,buf_rd_tmp
	; Now adjust the buffer
	movlw high(BUF_END+1)
	cpfseq FSR2H,ACCESS
	goto buf_read_no_wrap
	movlw low(BUF_END+1)
	cpfseq FSR2L,ACCESS
	goto buf_read_no_wrap
	movlw high(buf_start)
	movwf FSR2H,ACCESS
	movlw low(buf_start)
	movwf FSR2L,ACCESS
	; Previous read was missed
	; Read again from the right location
	movff INDF2,buf_rd_tmp
buf_read_no_wrap:
	; Now check the buffer empty condition
	banksel buf_wr_ptr
	movf buf_wr_ptr,W,BANKED
	cpfseq FSR2L,ACCESS
	goto buf_read_ok
	movf buf_wr_ptr+1,W,BANKED
	cpfseq FSR2H,ACCESS
	goto buf_read_ok
	; Set carry, return
	bsf STATUS,C,ACCESS
	return
buf_read_ok:
	; update the read pointer
	movff FSR2L,buf_rd_ptr
	movff FSR2H,buf_rd_ptr+1
	; clear carry, return
	movff buf_rd_tmp,WREG
	bcf STATUS,C,ACCESS
	return
END
