; 
; Author: Bradley A. Minch
; Organization: Franklin W. Olin College of Engineering
; Revision History: 
;     06/22/2005 - Added code to disable all endpoints on USRTIF and to mask
;                  bits 0, 1, and 7 of USTAT on TRNIF in ServiceUSB.
;     04/21/2005 - Initial public release.
;
; Later modifications by WZab (07.2008)
; Modifications introduced by Wojciech M. Zabolotny (wzab@ise.pw.edu.pl)
; Removed HID interface. Added communication via two endpoints (EP1 and EP2)
; to introduce my own protocol (used to communicate with the SD/MMC card).
; TODO: Sending long buffers

; ============================================================================
; 
; Peripheral Description:
; 
; This peripheral enumerates as a USB keyboard (an HID class device), which 
; continually sends the keycodes for 'f', 'o', 'o', 'b', 'a', 'r', and ' ' over 
; and over again.  It also has a CAPS LOCK key (a switch on RA0) and status LED 
; (an LED on RA1).  The firmware is currently configured to use an external 
; 4-MHz crystal, to operate as a low-speed device, and to use the internal 
; pull-up resistor.
;
;;; moving switch to RB4 (S2 on the fs picdem board)
;;; moving LED to RD1 (D2 on the fs picdem board)
; ============================================================================
;
; Software Licence Agreement:
; 
; THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION.  NO WARRANTIES, WHETHER 
; EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED 
; WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY 
; TO THIS SOFTWARE. THE AUTHOR SHALL NOT, UNDER ANY CIRCUMSTANCES, BE LIABLE 
; FOR SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
; 
	list      p=18f4550            ; list directive to define processor
	#include <p18f4550.inc>        ; processor specific variable definitions
	#include "usb_defs.inc"
	#include "ENGR2210.inc"
	#include "bit_bang_spi/mspi.inc"
		
#define TIMER0H_VAL         0xFE
#define TIMER0L_VAL         0x20

			PROCESSOR p18f4550

;;; TRY TO USE PICDEM'S CONFIG
;;; 
 			__CONFIG	_CONFIG1L, _PLLDIV_3_1L & _CPUDIV_OSC1_PLL2_1L & _USBDIV_2_1L 
; 			__CONFIG	_CONFIG1L, _PLLDIV_2_1L & _CPUDIV_OSC1_PLL2_1L & _USBDIV_2_1L 
; 			__CONFIG	_CONFIG1L, _PLLDIV_2_1L & _CPUDIV_OSC3_PLL4_1L & _USBDIV_2_1L
 			__CONFIG	_CONFIG1H, _FOSC_HSPLL_HS_1H & _FCMEN_OFF_1H & _IESO_OFF_1H
 			__CONFIG	_CONFIG2L, _PWRT_OFF_2L & _BOR_ON_2L & _BORV_2_2L & _VREGEN_ON_2L
 			__CONFIG	_CONFIG2H, _WDT_OFF_2H & _WDTPS_32768_2H
 			__CONFIG	_CONFIG3H, _MCLRE_ON_3H & _LPT1OSC_OFF_3H & _PBADEN_OFF_3H & _CCP2MX_ON_3H
 			__CONFIG	_CONFIG4L, _STVREN_ON_4L & _LVP_OFF_4L & _ICPRT_OFF_4L & _XINST_OFF_4L & _DEBUG_OFF_4L
				__CONFIG	_CONFIG5L, _CP0_OFF_5L & _CP1_OFF_5L & _CP2_OFF_5L & _CP3_OFF_5L
 			__CONFIG	_CONFIG5H, _CPB_OFF_5H & _CPD_OFF_5H
 			__CONFIG	_CONFIG6L, _WRT0_OFF_6L & _WRT1_OFF_6L & _WRT2_OFF_6L & _WRT3_OFF_6L
 			__CONFIG	_CONFIG6H, _WRTB_OFF_6H & _WRTC_OFF_6H & _WRTD_OFF_6H
 			__CONFIG	_CONFIG7L, _EBTR0_OFF_7L & _EBTR1_OFF_7L & _EBTR2_OFF_7L & _EBTR3_OFF_7L
 			__CONFIG	_CONFIG7H, _EBTRB_OFF_7H


;;; 1L:	F8 & F7 & FF = 0xf0
;;; 1H:	F2 & BF & 7F = 0x32
;;; 2L:	FE & FF & FF & FF = 0xfe
	

;;; #define	SHOW_ENUM_STATUS
	#include "bit_bang_spi/mspi_ext.inc"
	extern sd_init, sd_read_block, sd_address, sd_init_retry
	extern sd_write_init, sd_write_byte, sd_write_stop, sd_find_free
	extern sd_calc_max_block, sd_erase, sd_test_block
	extern long_delay
	extern ser_init, ser_rx_irq
	extern adc_init,adc_irq,adc_ptr,last_adc_ptr
	extern buf_init, buf_read, buf_write, buf_write_user, buf_wr_ptr
	extern LCD_INIT, LCD_INT, lcd_row_1, lcd_row_0
	extern DEL_N_MS

bank0		udata
USB_buffer_desc	res		4
USB_buffer_data	res		8
USB_error_flags	res 	1
USB_curr_config	res		1
USB_device_status	res	1
USB_protocol	res		1
USB_idle_rate	res		1
USB_dev_req	res			1
USB_address_pending	res	1
USB_desc_ptr	res		1
USB_bytes_left	res		1
USB_loop_index	res		1
USB_packet_length	res	1
USB_USTAT	res			1
USB_USWSTAT	res			1
COUNTER		res			1
LED_states	res			1
USB_LB_bytes_to_send res 2
USB_LB_ptr_to_send res 2
USB_LB_running res 1
Key_buffer	res			8
int_flag  res 1
my_cnt res 2
my_tmp res 1
fsr2_save res 2
sampling_flag res 1
	global sampling_flag
;STARTUP		code		0x2000
STARTUP		code		0x000
			goto		Main					; Reset vector
			nop
			nop
			call		HiIrq,FAST						; High-priority interrupt vector trap
			nop
			nop
			nop
			nop
			nop
			nop
			goto		$						; Low-priority interrupt vector trap

USBSTUFF	code		0x202a

Descriptor
			movlw		upper Descriptor_begin
			movwf		TBLPTRU, ACCESS
			movlw		high Descriptor_begin
			movwf		TBLPTRH, ACCESS
			movlw		low Descriptor_begin
			banksel		USB_desc_ptr
			addwf		USB_desc_ptr, W, BANKED
			ifset STATUS, C, ACCESS
				incf		TBLPTRH, F, ACCESS
				ifset STATUS, Z, ACCESS
					incf		TBLPTRU, F, ACCESS
				endi
			endi
			movwf		TBLPTRL, ACCESS
			tblrd*
			movf		TABLAT, W
			return

Descriptor_begin
Device
			db			0x12, DEVICE		; bLength, bDescriptorType
			db			0x00, 0x02			; bcdUSB (low byte), bcdUSB (high byte)
			db			0xff, 0x00			; bDeviceClass, bDeviceSubClass
			db			0xff, 0x08			; bDeviceProtocl, bMaxPacketSize
			db			0xa5, 0xa5			; idVendor (low byte), idVendor (high byte)
			db			0x02, 0x01			; idProduct (low byte), idProduct (high byte)
			db			0x01, 0x00			; bcdDevice (low byte), bcdDevice (high byte)
			db			0x01, 0x02			; iManufacturer, iProduct
;;;			db			0x00, 0x00			; iManufacturer, iProduct
			db			0x00, NUM_CONFIGURATIONS	; iSerialNumber (none), bNumConfigurations

Configuration1
			db			0x09, CONFIGURATION	; bLength, bDescriptorType
			db			0x20, 0x00			; wTotalLength (low byte), wTotalLength (high byte)
			db			NUM_INTERFACES, 0x01	; bNumInterfaces, bConfigurationValue
			db			0x00, 0xA0			; iConfiguration (none), bmAttributes
			db			0xa0, 0x09 ; bMaxPower (100 mA), bLength (Interface1 descriptor starts here)
			db			INTERFACE, 0x00		; bDescriptorType, bInterfaceNumber
			db			0x00, 0x02			; bAlternateSetting, bNumEndpoints (excluding EP0)
			db			0xff, 0xff			; bInterfaceClass (HID code), bInterfaceSubClass (Boot subclass)
			db			0xff, 0x00			; bInterfaceProtocol (Keyboard protocol), iInterface (none)
; Start of endpoint 1 descriptor
			db			0x07, ENDPOINT
			db			0x1, 2; EP1 IN, 
			db			0x40, 0; transfer size
			db 		0, 0x07 ; Start of enpoint 2 descriptor
			db			ENDPOINT,0x82
			db			2, 0x40 ; EP1 IN, transfer size
			db 		0, 0

; Start of endpoint 2 descriptor

String0
			db			String1-String0, STRING	; bLength, bDescriptorType
			db			0x09, 0x04			; wLANGID[0] (low byte), wLANGID[0] (high byte)
String1
			db			String2-String1, STRING	; bLength, bDescriptorType
			db			'W', 0x00			; bString
			db			'Z', 0x00
			db			'a', 0x00
			db			'b', 0x00
			db			' ', 0x00
			db			'S', 0x00
			db			'o', 0x00
			db			'f', 0x00
			db			't', 0x00
			db			'w', 0x00
			db			'a', 0x00
			db			'r', 0x00
			db			'e', 0x00
String2
			db			Descriptor_end-String2, STRING	; bLength, bDescriptorType
			db			'G', 0x00			; bString
			db			'P', 0x00
			db			'S', 0x00
			db			' ', 0x00
			db			'l', 0x00
			db			'o', 0x00
			db			'g', 0x00
			db			'g', 0x00
			db			'e', 0x00
			db			'r', 0x00
			db			' ', 0x00
			db			'V', 0x00
			db			'1', 0x00
			db			'.', 0x00
			db			'1', 0x00
Descriptor_end

InitUSB
		clrf	UIE, ACCESS				; mask all USB interrupts
		clrf	UIR, ACCESS				; clear all USB interrupt flags
;;; 		movlw	0x10
		movlw	0x14
		movwf	UCFG, ACCESS			; configure USB for low-speed transfers and to use the on-chip transciever and pull-up resistor
		movlw	0x08
		movwf	UCON, ACCESS			; enable the USB module and its supporting circuitry
		banksel	USB_curr_config
		clrf	USB_curr_config, BANKED
		clrf	USB_idle_rate, BANKED
		clrf	USB_USWSTAT, BANKED		; default to powered state
		movlw	0x01
		movwf	USB_device_status, BANKED
		movwf	USB_protocol, BANKED	; default protocol to report protocol initially
		movlw	NO_REQUEST
		movwf	USB_dev_req, BANKED		; No device requests in process
#ifdef SHOW_ENUM_STATUS
		movlw	b'11110000'		; RD0-3 outputs (LEDs 1-4), RD4-7 inputs
		movwf	TRISD, ACCESS
		movlw	0x01
		movwf	PORTD, ACCESS			; set bit zero to indicate Powered status
#endif
		return

ServiceUSB
		select
			caseset	UIR, UERRIF, ACCESS
				clrf		UEIR, ACCESS
				break
			caseset UIR, SOFIF, ACCESS
				bcf			UIR, SOFIF, ACCESS
				break
			caseset	UIR, IDLEIF, ACCESS
				bcf			UIR, IDLEIF, ACCESS
;;; 				bsf			UCON, SUSPND, ACCESS
#ifdef SHOW_ENUM_STATUS
;;; 				movlw		0xE0
;;; 				andwf		PORTB, F, ACCESS
;;; 				bsf			PORTB, 4, ACCESS
#endif
				break
			caseset UIR, ACTVIF, ACCESS
				bcf			UIR, ACTVIF, ACCESS
				bcf			UCON, SUSPND, ACCESS
#ifdef SHOW_ENUM_STATUS
				movlw		0xF0
				andwf		PORTD, F, ACCESS
				banksel		USB_USWSTAT
				movf		USB_USWSTAT, W, BANKED
				select
					case POWERED_STATE
						movlw	0x01
						break
					case DEFAULT_STATE
						movlw	0x02
						break
					case ADDRESS_STATE
						movlw	0x04
						break
					case CONFIG_STATE
						movlw	0x08
				ends
				iorwf		PORTD, F, ACCESS
#endif
				break
			caseset	UIR, STALLIF, ACCESS
				bcf			UIR, STALLIF, ACCESS
				break
			caseset	UIR, URSTIF, ACCESS
				; bsf		PORTD, 1, ACCESS
				banksel		USB_curr_config
				clrf		USB_curr_config, BANKED
				bcf			UIR, TRNIF, ACCESS		; clear TRNIF four times to clear out the USTAT FIFO
				bcf 		UIR, TRNIF, ACCESS
				bcf			UIR, TRNIF, ACCESS
				bcf			UIR, TRNIF, ACCESS
				clrf		UEP0, ACCESS			; clear all EP control registers to disable all endpoints
				clrf		UEP1, ACCESS
				clrf		UEP2, ACCESS
				clrf		UEP3, ACCESS
				clrf		UEP4, ACCESS
				clrf		UEP5, ACCESS
				clrf		UEP6, ACCESS
				clrf		UEP7, ACCESS
				clrf		UEP8, ACCESS
				clrf		UEP9, ACCESS
				clrf		UEP10, ACCESS
				clrf		UEP11, ACCESS
				clrf		UEP12, ACCESS
				clrf		UEP13, ACCESS
				clrf		UEP14, ACCESS
				clrf		UEP15, ACCESS
				banksel		BD0OBC
				movlw		0x08
				movwf		BD0OBC, BANKED
				movlw		low USB_Buffer			; EP0 OUT gets a buffer...
				movwf		BD0OAL, BANKED
				movlw		high USB_Buffer
				movwf		BD0OAH, BANKED			; ...set up its address
				movlw		0x88					; set UOWN bit (USB can write)
				movwf		BD0OST, BANKED
				movlw		low (USB_Buffer+0x08)	; EP0 IN gets a buffer...
				movwf		BD0IAL, BANKED
				movlw		high (USB_Buffer+0x08)
				movwf		BD0IAH, BANKED			; ...set up its address
				movlw		0x08					; clear UOWN bit (MCU can write)
				movwf		BD0IST, BANKED
				clrf		UADDR, ACCESS			; set USB Address to 0
 				clrf		UIR, ACCESS				; clear all the USB interrupt flags
				movlw		ENDPT_CONTROL
				movwf		UEP0, ACCESS			; EP0 is a control pipe and requires an ACK
 				movlw		0xFF					; enable all error interrupts
 				movwf		UEIE, ACCESS
				banksel		USB_USWSTAT
				movlw		DEFAULT_STATE
				movwf		USB_USWSTAT, BANKED
				movlw		0x01
				movwf		USB_device_status, BANKED	; self powered, remote wakeup disabled
#ifdef SHOW_ENUM_STATUS
				movlw		0xF0
				andwf		PORTD, F, ACCESS
				bsf 		PORTD, 1, ACCESS		; set bit 1 of PORTB to indicate Powered state
#endif
				; bsf		PORTD, 2, ACCESS
		
				break
			caseset	UIR, TRNIF, ACCESS
				; bsf		PORTD, 3, ACCESS
				movlw		0x04
				movwf		FSR0H, ACCESS
				movf		USTAT, W, ACCESS
				andlw		0x7C					; mask out bits 0, 1, and 7 of USTAT
				movwf		FSR0L, ACCESS
				banksel		USB_buffer_desc
				movf		POSTINC0, W
				movwf		USB_buffer_desc, BANKED
				movf		POSTINC0, W
				movwf		USB_buffer_desc+1, BANKED
				movf		POSTINC0, W
				movwf		USB_buffer_desc+2, BANKED
				movf		POSTINC0, W
				movwf		USB_buffer_desc+3, BANKED
				movf		USTAT, W, ACCESS
				movwf		USB_USTAT, BANKED		; save the USB status register
				bcf			UIR, TRNIF, ACCESS		; clear TRNIF interrupt flag
#ifdef SHOW_ENUM_STATUS
				andlw		0x18					; extract EP bits
				select
					case EP0
						movlw		0x20
						break
					case EP1
						movlw		0x40
						break
					case EP2
						movlw		0x80
						break
				ends
;;; 					xorwf		PORTD, F, ACCESS		; toggle bit 5, 6, or 7 of PORTD to reflect EP activity
#endif
				movf		USB_buffer_desc, W, BANKED
				andlw		0x3C					; extract PID bits
				select
					case TOKEN_SETUP
						call		ProcessSetupToken
						break
					case TOKEN_IN
						call		ProcessInToken
						break
					case TOKEN_OUT
						call		ProcessOutToken
						break
				ends
				break
		ends
		return

ProcessSetupToken
			banksel		USB_buffer_data
			movf		USB_buffer_desc+ADDRESSH, W, BANKED
			movwf		FSR0H, ACCESS
			movf		USB_buffer_desc+ADDRESSL, W, BANKED
			movwf		FSR0L, ACCESS
			movf		POSTINC0, W
			movwf		USB_buffer_data, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+1, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+2, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+3, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+4, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+5, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+6, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+7, BANKED
			banksel		BD0OBC
			movlw		0x08
			movwf		BD0OBC, BANKED					; reset the byte count
			movwf		BD0IST, BANKED					; return the in buffer to us (dequeue any pending requests)
			banksel		USB_buffer_data+bmRequestType
			ifl	USB_buffer_data+bmRequestType, EQ, 0x21
				movlw		0xC8
			otherwise
				movlw		0x88
			endi
			banksel		BD0OST
			movwf		BD0OST, BANKED					; set EP0 OUT UOWN back to USB and DATA0/DATA1 packet according to request type
			bcf			UCON, PKTDIS, ACCESS			; assuming there is nothing to dequeue, clear the packet disable bit
			banksel		USB_dev_req
			movlw		NO_REQUEST
			movwf		USB_dev_req, BANKED				; clear the device request in process
			movf		USB_buffer_data+bmRequestType, W, BANKED
			andlw		0x60							; extract request type bits
			select
				case STANDARD
					call		StandardRequests
					break
				case CLASS
					call		ClassRequests
					break
				case VENDOR
					call		VendorRequests
					break
				default
					bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
			ends
			return

StandardRequests
			movf		USB_buffer_data+bRequest, W, BANKED
			select
				case GET_STATUS
					movf		USB_buffer_data+bmRequestType, W, BANKED
					andlw		0x1F					; extract request recipient bits
					select
						case RECIPIENT_DEVICE
							banksel		BD0IAH
							movf		BD0IAH, W, BANKED
							movwf		FSR0H, ACCESS
							movf		BD0IAL, W, BANKED				; get buffer pointer
							movwf		FSR0L, ACCESS
							banksel		USB_device_status
							movf		USB_device_status, W, BANKED	; copy device status byte to EP0 buffer
							movwf		POSTINC0
							clrf		INDF0
							banksel		BD0IBC
							movlw		0x02
							movwf		BD0IBC, BANKED					; set byte count to 2
							movlw		0xC8
							movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
							break
						case RECIPIENT_INTERFACE
							movf		USB_USWSTAT, W, BANKED
							select
								case ADDRESS_STATE
									bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
									break
								case CONFIG_STATE
									ifl USB_buffer_data+wIndex, LT, NUM_INTERFACES
										banksel		BD0IAH
										movf		BD0IAH, W, BANKED
										movwf		FSR0H, ACCESS
										movf		BD0IAL, W, BANKED				; get buffer pointer
										movwf		FSR0L, ACCESS
										clrf		POSTINC0
										clrf		INDF0
										movlw		0x02
										movwf		BD0IBC, BANKED					; set byte count to 2
										movlw		0xC8
										movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit																								
									otherwise
										bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
									endi
									break
							ends
							break
						case RECIPIENT_ENDPOINT
							movf		USB_USWSTAT, W, BANKED
							select
								case ADDRESS_STATE
									movf		USB_buffer_data+wIndex, W, BANKED	; get EP
									andlw		0x0F								; strip off direction bit
									ifset STATUS, Z, ACCESS							; see if it is EP0
										banksel		BD0IAH
										movf		BD0IAH, W, BANKED				; put EP0 IN buffer pointer...
										movwf		FSR0H, ACCESS
										movf		BD0IAL, W, BANKED
										movwf		FSR0L, ACCESS					; ...into FSR0
										ifset UEP0, EPSTALL, ACCESS
											movlw		0x01
										otherwise
											movlw		0x00
										endi
										movwf		POSTINC0
										clrf		INDF0
										movlw		0x02
										movwf		BD0IBC, BANKED					; set byte count to 2
										movlw		0xC8
										movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
									otherwise
										bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
									endi
									break
								case CONFIG_STATE
									banksel		BD0IAH
									movf		BD0IAH, W, BANKED					; put EP0 IN buffer pointer...
									movwf		FSR0H, ACCESS
									movf		BD0IAL, W, BANKED
									movwf		FSR0L, ACCESS						; ...into FSR0
									movlw		high UEP0							; put UEP0 address...
									movwf		FSR1H, ACCESS
									movlw		low UEP0
									movwf		FSR1L, ACCESS						; ...into FSR1
									banksel		USB_buffer_data+wIndex
									movf		USB_buffer_data+wIndex, W, BANKED	; get EP and...
									andlw		0x0F								; ...strip off direction bit
									ifclr PLUSW1, EPOUTEN, ACCESS
									andifclr PLUSW1, EPINEN, ACCESS
										bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
									otherwise
										ifset PLUSW1, EPSTALL, ACCESS
											movlw		0x01
										otherwise
											movlw		0x00
										endi
										movwf		POSTINC0
										clrf		INDF0
										banksel		BD0IBC
										movlw		0x02
										movwf		BD0IBC, BANKED					; set byte count to 2
										movlw		0xC8
										movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
									endi
									break
								default
									bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							ends
							break
						default
							bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					ends
					break
				case CLEAR_FEATURE
				case SET_FEATURE
					movf		USB_buffer_data+bmRequestType, W, BANKED
					andlw		0x1F					; extract request recipient bits
					select
						case RECIPIENT_DEVICE
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case DEVICE_REMOTE_WAKEUP
									ifl USB_buffer_data+bRequest, EQ, CLEAR_FEATURE
										bcf			USB_device_status, 1, BANKED
									otherwise
										bsf			USB_device_status, 1, BANKED
									endi
									banksel		BD0IBC
									clrf		BD0IBC, BANKED					; set byte count to 0
									movlw		0xC8
									movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
									break
								default
									bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
							ends
							break
						case RECIPIENT_ENDPOINT
							movf		USB_USWSTAT, W, BANKED
							select
								case ADDRESS_STATE
									movf		USB_buffer_data+wIndex, W, BANKED	; get EP
									andlw		0x0F								; strip off direction bit
									ifset STATUS, Z, ACCESS							; see if it is EP0
										ifl USB_buffer_data+bRequest, EQ, CLEAR_FEATURE
											bcf			UEP0, EPSTALL, ACCESS
										otherwise
											bsf			UEP0, EPSTALL, ACCESS
										endi
										banksel		BD0IBC
										clrf		BD0IBC, BANKED					; set byte count to 0
										movlw		0xC8
										movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
									otherwise
										bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
									endi
									break
								case CONFIG_STATE
									movlw		high UEP0							; put UEP0 address...
									movwf		FSR0H, ACCESS
									movlw		low UEP0
									movwf		FSR0L, ACCESS						; ...into FSR0
									movf		USB_buffer_data+wIndex, W, BANKED	; get EP
									andlw		0x0F								; strip off direction bit
									addwf		FSR0L, F, ACCESS					; add EP number to FSR0
									ifset		STATUS, C, ACCESS
										incf		FSR0H, F, ACCESS
									endi
									ifclr INDF0, EPOUTEN, ACCESS
									andifclr INDF0, EPINEN, ACCESS
										bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
									otherwise
										ifl USB_buffer_data+bRequest, EQ, CLEAR_FEATURE
											bcf			INDF0, EPSTALL, ACCESS
										otherwise
											bsf			INDF0, EPSTALL, ACCESS
										endi
										banksel		BD0IBC
										clrf		BD0IBC, BANKED			; set byte count to 0
										movlw		0xC8
										movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
									endi
									break
								default
									bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							ends
							break
						default
							bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					ends
					break
				case SET_ADDRESS
					ifset USB_buffer_data+wValue, 7, BANKED		; if new device address is illegal, send Request Error
						bsf			UEP0, EPSTALL, ACCESS		; set EP0 protocol stall bit to signify Request Error
					otherwise
						movlw		SET_ADDRESS
						movwf		USB_dev_req, BANKED			; processing a SET_ADDRESS request
						movf		USB_buffer_data+wValue, W, BANKED
						movwf		USB_address_pending, BANKED	; save new address
						banksel		BD0IBC
						clrf		BD0IBC, BANKED				; set byte count to 0
						movlw		0xC8
						movwf		BD0IST, BANKED				; send packet as DATA1, set UOWN bit
					endi
					break
				case GET_DESCRIPTOR
					movwf		USB_dev_req, BANKED				; processing a GET_DESCRIPTOR request
					movf		USB_buffer_data+(wValue+1), W, BANKED
					select
						case DEVICE
							movlw		low (Device-Descriptor_begin)
							movwf		USB_desc_ptr, BANKED
							call		Descriptor				; get descriptor length
							movwf		USB_bytes_left, BANKED
							ifl USB_buffer_data+(wLength+1), EQ, 0
							andiff USB_buffer_data+wLength, LT, USB_bytes_left
								movf		USB_buffer_data+wLength, W, BANKED
								movwf		USB_bytes_left, BANKED
							endi
							call		SendDescriptorPacket
							break
						case CONFIGURATION
							bcf			USB_error_flags, 0, BANKED
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case 0
									movlw		low (Configuration1-Descriptor_begin)
									break
								default
									bsf			USB_error_flags, 0, BANKED
							ends
							ifclr USB_error_flags, 0, BANKED
								addlw		0x02				; add offset for wTotalLength
								movwf		USB_desc_ptr, BANKED
								call		Descriptor			; get total descriptor length
								movwf		USB_bytes_left, BANKED
								movlw		0x02
								subwf		USB_desc_ptr, F, BANKED	; subtract offset for wTotalLength
								ifl USB_buffer_data+(wLength+1), EQ, 0
								andiff USB_buffer_data+wLength, LT, USB_bytes_left
									movf		USB_buffer_data+wLength, W, BANKED
									movwf		USB_bytes_left, BANKED
								endi
								call		SendDescriptorPacket
							otherwise
								bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							endi
							break
						case STRING
							bcf			USB_error_flags, 0, BANKED
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case 0
									movlw		low (String0-Descriptor_begin)
									break
								case 1
									movlw		low (String1-Descriptor_begin)
									break
								case 2
									movlw		low (String2-Descriptor_begin)
									break
								default
									bsf			USB_error_flags, 0, BANKED
							ends
							ifclr USB_error_flags, 0, BANKED
								movwf		USB_desc_ptr, BANKED
								call		Descriptor		; get descriptor length
								movwf		USB_bytes_left, BANKED
								ifl USB_buffer_data+(wLength+1), EQ, 0
								andiff USB_buffer_data+wLength, LT, USB_bytes_left
									movf		USB_buffer_data+wLength, W, BANKED
									movwf		USB_bytes_left, BANKED
								endi
								call		SendDescriptorPacket
							otherwise
								bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							endi
							break
						default
							bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					ends
					break
				case GET_CONFIGURATION
					banksel		BD0IAH
					movf		BD0IAH, W, BANKED
					movwf		FSR0H, ACCESS
					movf		BD0IAL, W, BANKED
					movwf		FSR0L, ACCESS
					banksel		USB_curr_config
					movf		USB_curr_config, W, BANKED
					movwf		INDF0					; copy current device configuration to EP0 IN buffer
					banksel		BD0IBC
					movlw		0x01
					movwf		BD0IBC, BANKED			; set EP0 IN byte count to 1
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				case SET_CONFIGURATION
					ifl USB_buffer_data+wValue, LE, NUM_CONFIGURATIONS
						movf		USB_buffer_data+wValue, W, BANKED
						movwf		USB_curr_config, BANKED
						select
							case 0
								movlw		ADDRESS_STATE
								movwf		USB_USWSTAT, BANKED
#ifdef SHOW_ENUM_STATUS
								movlw		0xF0
								andwf		PORTD, F, ACCESS
								bsf			PORTD, 2, ACCESS
#endif
						break
							default
								movlw		CONFIG_STATE
								movwf		USB_USWSTAT, BANKED
								movlw		0x040
								banksel		BD1OBC
								movwf		BD1OBC, BANKED			; set EP1 IN byte count to 64
								movlw		low (USB_Buffer+0x10)
								movwf		BD1OAL, BANKED			; set EP1 IN buffer address
								movlw		high (USB_Buffer+0x10)
								movwf		BD1OAH, BANKED
								movlw		0x040
								banksel		BD2IBC
								movwf		BD2IBC, BANKED			; set EP2 IN byte count to 64
								movlw		low (USB_Buffer2)
								movwf		BD2IAL, BANKED			; set EP1 IN buffer address
								movlw		high (USB_Buffer2)
								movwf		BD2IAH, BANKED
								movlw		0x48
								movwf		BD2IST, BANKED			; clear UOWN bit (PIC can write EP2 IN buffer)
								movlw		0x88
								banksel	BD1OST
								movwf		BD1OST, BANKED			; set UOWN bit (USB can write EP1 OUT buffer)
								movlw		ENDPT_OUT_ONLY
								movwf		UEP1, ACCESS			; enable EP1
								movlw		ENDPT_IN_ONLY
								movwf		UEP2, ACCESS			; enable EP2
#ifdef SHOW_ENUM_STATUS
								movlw		0xF0
								andwf		PORTD, F, ACCESS
								bsf			PORTD, 3, ACCESS
#endif
						ends
						banksel		BD0IBC
						clrf		BD0IBC, BANKED			; set byte count to 0
						movlw		0xC8
						movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					otherwise
						bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					endi
					break
				case GET_INTERFACE
					movf		USB_USWSTAT, W, BANKED
					select
						case CONFIG_STATE
							ifl USB_buffer_data+wIndex, LT, NUM_INTERFACES
								banksel		BD0IAH
								movf		BD0IAH, W, BANKED
								movwf		FSR0H, ACCESS
								movf		BD0IAL, W, BANKED		; get buffer pointer
								movwf		FSR0L, ACCESS
								clrf		INDF0					; always send back 0 for bAlternateSetting
								movlw		0x01
								movwf		BD0IBC, BANKED			; set byte count to 1
								movlw		0xC8
								movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
							otherwise
								bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							endi
							break
						default
							bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					ends
					break
				case SET_INTERFACE
					movf		USB_USWSTAT, W, BANKED
					select
						case CONFIG_STATE
							ifl USB_buffer_data+wIndex, LT, NUM_INTERFACES
								movf		USB_buffer_data+wValue, W, BANKED
								select
									case 0									; currently support only bAlternateSetting of 0
										banksel		BD0IBC
										clrf		BD0IBC, BANKED			; set byte count to 0
										movlw		0xC8
										movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
										break
									default
										bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
								ends
							otherwise
								bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							endi
							break
						default
							bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					ends
					break
				case SET_DESCRIPTOR
				case SYNCH_FRAME
				default
					bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					break
			ends
			return

VendorRequests
		movf		USB_buffer_data+bRequest, W, BANKED
		select
			default
				bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
		ends
		return

ClassRequests
		movf		USB_buffer_data+bRequest, W, BANKED
		select
			default
				bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
		ends
		return

ProcessInToken
			banksel		USB_USTAT
			movf		USB_USTAT, W, BANKED
			andlw		0x18		; extract the EP bits
			select
				case EP0
					movf		USB_dev_req, W, BANKED
					select
						case SET_ADDRESS
							movf		USB_address_pending, W, BANKED
							movwf		UADDR, ACCESS
							select
								case 0
									movlw		DEFAULT_STATE
									movwf		USB_USWSTAT, BANKED
#ifdef SHOW_ENUM_STATUS
									movlw		0xF0
									andwf		PORTD, F, ACCESS
									bsf			PORTD, 1, ACCESS
#endif
									break
								default
									movlw		ADDRESS_STATE
									movwf		USB_USWSTAT, BANKED
#ifdef SHOW_ENUM_STATUS
									movlw		0xF0
									andwf		PORTD, F, ACCESS
									bsf			PORTD, 2, ACCESS
#endif
							ends
							break
						case GET_DESCRIPTOR
							call		SendDescriptorPacket
							break
					ends
					break
				case EP1
					break
				case EP2
					call SendLongBuffer
					break
			ends
			return

SendW
			movwf POSTINC1, ACCESS
			return

SendOK
			movlw 'O'
			call SendW
			movlw 'K'
			call SendW
			return

ProcessOutToken
			banksel		USB_USTAT
			movf		USB_USTAT, W, BANKED
			andlw		0x18		; extract the EP bits
			select
				case EP0
					movf		USB_dev_req, W, BANKED
					select
						case SET_REPORT
							movlw		NO_REQUEST
							movwf		USB_dev_req, BANKED	; clear device request
							banksel		BD0OAH
							movf		BD0OAH, W, BANKED	; put EP0 OUT buffer pointer...
							movwf		FSR0H, ACCESS
							movf		BD0OAL, W, BANKED
							movwf		FSR0L, ACCESS		; ...into FSR0
							movf		INDF0, W			; get the first byte in the buffer and...
							banksel		LED_states
							movwf		LED_states, BANKED	; ...update the LED states with it
;;; 							ifset LED_states, 1, BANKED		; set RA1 to reflect the new state of the CAPS LOCK LED
;;; 								bsf			PORTD, 0, ACCESS
;;; 							otherwise
;;; 								bcf			PORTD, 0, ACCESS
;;; 							endi
					ends
					banksel		BD0OBC
					movlw		0x08
					movwf		BD0OBC, BANKED
					movlw		0x88
					movwf		BD0OST, BANKED
					clrf		BD0IBC, BANKED		; set byte count to 0
					movlw		0xC8
					movwf		BD0IST, BANKED		; send packet as DATA1, set UOWN bit
					break
				case EP1
					banksel	USB_dev_req
					movlw		NO_REQUEST
					movwf		USB_dev_req, BANKED	; clear device request
					; Setup the input buffer
					movff		BD1OAL,FSR0L
					movff		BD1OAH,FSR0H
					; Setup the output buffer
					movlw		low(USB_Buffer2)
					movwf		FSR1L, ACCESS
					movlw		high(USB_Buffer2)
					movwf		FSR1H, ACCESS
					; Now check the command
					movf		POSTINC0, W, ACCESS
					select
						case 'S'
							bsf cs_port, cs_pin, ACCESS
							call SendOK
							break
						case 'D'
							bcf cs_port, cs_pin, ACCESS
							call SendOK
							break
						case 'I'
							; Send the data and receive the response
							movlw 'R'
							call SendW
							banksel	BD1OBC
							decf		BD1OBC,F,BANKED
							; Now loop over the sent bytes
loop_wz_12:				
								movf	POSTINC0, W, ACCESS
								call	mspi_byte_wio
								call	SendW
								banksel BD1OBC
								decfsz	BD1OBC, F
								goto		loop_wz_12
							break
						case 'M'
							; Initialize the card
							call sd_init
							break
						case 'T'
							; Retry to initialize the card
							call sd_init_retry
							break
						case 'W'
							call long_delay
							movlw 'O'
							call SendW
							movlw 'K'
							call SendW
							break
						case 'P'
							banksel COUNTER
							movf POSTINC0,W,ACCESS
							movwf COUNTER,BANKED
							movf POSTINC0,W,ACCESS
							movwf FSR0L,ACCESS
							movlw 0x0f
							movwf FSR0H,ACCESS
							movf COUNTER,W,BANKED
							movwf INDF0,ACCESS
							movlw 'O'
							call SendW
							movlw 'K'
							call SendW
							break
						case 'E'
							movf POSTINC0,W,ACCESS
							movwf FSR0L,ACCESS
							movlw 0x0f
							movwf FSR0H,ACCESS
							movf INDF0,W,ACCESS
							call SendW
							break
						case 'X' ; Sample the RS232 signal!
							clrf RCSTA,ACCESS
							movf POSTINC0,W,ACCESS
							banksel my_cnt
							movwf my_cnt, BANKED
							movf POSTINC0,W,ACCESS
							movwf my_cnt+1, BANKED
							banksel COUNTER
							clrf COUNTER,BANKED
							; Wait for 0 on RX line
x_wait_0:				btfsc PORTC,6,ACCESS
							goto x_wait_0
x_wait_1:				btfss PORTC,6,ACCESS
							goto x_wait_1
							;Sampling
x_wait_2:
							movf PORTC,W,ACCESS
							andlw 0x40
							movwf POSTINC1,ACCESS
							movff my_cnt+1,my_tmp
x_wait_3a:
							banksel my_cnt							
							movf my_cnt,W,BANKED
x_wait_3:
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							nop
							decfsz WREG,ACCESS
							goto x_wait_3
							banksel my_tmp
							decfsz my_tmp,F,BANKED
							goto x_wait_3a
							banksel COUNTER
							decfsz COUNTER,BANKED
							goto x_wait_2
							break
						case 'Y'
							; Read the GPS&ADC samples, when they are ready
							banksel COUNTER
							movlw   0x80
							movwf	  COUNTER,BANKED
adc_wait_0:
							call buf_read
							bc adc_wait_0
							movwf POSTINC1,ACCESS
							banksel COUNTER
							decfsz COUNTER,F,BANKED
							goto adc_wait_0
							break
						case 'R'
							; Read the block from the card
							; Read the requested address
							banksel sd_address
							movf POSTINC0,W,ACCESS
							movwf sd_address, BANKED
							movf POSTINC0,W,ACCESS
							movwf sd_address+1, BANKED
							movf POSTINC0,W,ACCESS
							movwf sd_address+2, BANKED
							movf POSTINC0,W,ACCESS
							movwf sd_address+3, BANKED
							call sd_read_block
							;check the result
							select
								case 0
									nop
									break
								default
									call SendW
									movlw 'N'
									call SendW
									movlw 'O'
									call SendW
								break
							ends
							break
						case 'A'
							; Start writing to the card
							; Read the requested address
							banksel sd_address
							movf POSTINC0,W,ACCESS
							movwf sd_address+0, BANKED
							movf POSTINC0,W,ACCESS
							movwf sd_address+1, BANKED
							movf POSTINC0,W,ACCESS
							movwf sd_address+2, BANKED
							movf POSTINC0,W,ACCESS
							movwf sd_address+3, BANKED
							call sd_write_init
							movwf POSTINC1,BANKED
							;check the result
							select
								case 0
									nop
									break
								default
									call SendW
									movlw 'N'
									call SendW
									movlw 'O'
									call SendW
								break
							ends
							break
						case 'B'
							; Send the data
							banksel	BD1OBC
							decf		BD1OBC,F,BANKED
							; Now loop over the sent bytes
loop_wz_13:				
								movf	POSTINC0, W, ACCESS
								call	sd_write_byte
								btfsc STATUS,C,ACCESS
								goto b_cmd_error
								banksel BD1OBC
								decfsz	BD1OBC, F
								goto		loop_wz_13
							movlw 'O'
							call SendW
							movlw 'K'
							call SendW
							goto b_cmd_exit
b_cmd_error:
							call SendW
							movlw 'N'
							call SendW
							movlw 'O'
							call SendW
b_cmd_exit:
							break
						case 'C'
							; Stop writing to the card
							call sd_write_stop
							call SendW
							movlw 'O'
							call SendW
							movlw 'K'
							call SendW
							break
						case 'K'
							; Set the baudrate
							movf POSTINC0,W,ACCESS
							movwf SPBRGH, ACCESS
							movf POSTINC0,W,ACCESS
							movwf SPBRG, ACCESS
							; Set the BAUDCON
							movf POSTINC0,W,ACCESS
							movwf BAUDCON, ACCESS
							; Set the TXSTA
							movf POSTINC0,W,ACCESS
							movwf TXSTA, ACCESS
							movlw 'O'
							call SendW
							movlw 'K'
							call SendW
							break
						case 'G'
							banksel int_flag
							movf int_flag,W,BANKED
							movwf POSTINC1,ACCESS
							movf RCREG,W,ACCESS
							movf RCREG,W,ACCESS
							movf RCREG,W,ACCESS
							movf RCREG,W,ACCESS
							movf RCREG,W,ACCESS
							movf RCREG,W,ACCESS
							movf RCREG,W,ACCESS
							movf RCON,W,ACCESS
							movwf POSTINC1,ACCESS
							movf STKPTR,W,ACCESS
							movwf POSTINC1,ACCESS
							movf INTCON,W,ACCESS
							movwf POSTINC1,ACCESS
							bcf RCSTA,CREN,ACCESS
							bsf RCSTA,CREN,ACCESS
							movf RCSTA,W,ACCESS
							movwf POSTINC1,ACCESS
							movf PIR2,W,ACCESS
							movwf POSTINC1,ACCESS
							movf PIR1,W,ACCESS
							movwf POSTINC1,ACCESS
							movf PIE1,W,ACCESS
							movwf POSTINC1,ACCESS
							movf RCREG,W,ACCESS
							movwf POSTINC1,ACCESS
							movlw 'O'
							call SendW
							movlw 'K'
							call SendW
							break
						case 'H'
							banksel int_flag
							movf int_flag,W,BANKED
							movwf POSTINC1,ACCESS
							movf FSR2L,W,ACCESS
							movwf POSTINC1,ACCESS
							movf INTCON,W,ACCESS
							movwf POSTINC1,ACCESS
							movf RCSTA,W,ACCESS
							movwf POSTINC1,ACCESS
							movf PIR1,W,ACCESS
							movwf POSTINC1,ACCESS
							movf PIE1,W,ACCESS
							movwf POSTINC1,ACCESS
							movf RCREG,W,ACCESS
							movwf POSTINC1,ACCESS
							banksel COUNTER
							clrf COUNTER,BANKED
							lfsr 0,0x300
loop_wz_14:
							movf POSTINC0,W,ACCESS
							movwf POSTINC1,ACCESS
							decfsz COUNTER,F,BANKED
							goto loop_wz_14
							break
						case 'F'
							; Start data acquisition from the block "0"
							banksel sd_address
							clrf sd_address,BANKED
							clrf sd_address+1,BANKED
							clrf sd_address+2,BANKED
							clrf sd_address+3,BANKED
							call sd_write_init
							; Check the result
							xorlw 0x0
							bnz f_cmd_error
							; Set the sampling flag...
							banksel sampling_flag
							setf sampling_flag, BANKED
							movlw 'O'
							call SendW
							movlw 'K'
							call SendW
							goto f_cmd_exit
f_cmd_error:
							call SendW
							movlw 'N'
							call SendW
							movlw 'O'
							call SendW
f_cmd_exit:
							break
						case 'J'
							banksel sampling_flag
							clrf sampling_flag, BANKED
							; empty the buffer
							call write_samples
							call sd_write_stop
							bc j_cmd_error
							movlw 'O'
							call SendW
							movlw 'K'
							call SendW
							goto j_cmd_exit
j_cmd_error:
							call SendW
							movlw 'N'
							call SendW
							movlw 'O'
							call SendW
j_cmd_exit:
							break
						case 'Q'
							call sd_erase
							xorlw 0
							bnz q_cmd_error
							movlw 'O'
							call SendW
							movlw 'K'
							call SendW
							goto q_cmd_exit
q_cmd_error:
							call SendW
							movlw 'N'
							call SendW
							movlw 'O'
							call SendW
q_cmd_exit:
							break
						default
							movlw 'N'
							call SendW
							movlw 'O'
							call SendW
							break
					ends
					movff FSR1L, USB_LB_bytes_to_send
					movff FSR1H, USB_LB_bytes_to_send+1
					banksel USB_LB_bytes_to_send
					movlw low(USB_Buffer2)
					subwf USB_LB_bytes_to_send,F,BANKED
					movlw high(USB_Buffer2)
					subwfb USB_LB_bytes_to_send+1,F,BANKED
					movlw		low(USB_Buffer2)
					banksel	USB_LB_ptr_to_send
					movwf		USB_LB_ptr_to_send
					movlw		high(USB_Buffer2)
					movwf		USB_LB_ptr_to_send+1
					banksel	USB_LB_running
					setf 		USB_LB_running
					banksel	BD1OBC
					movlw		0x40
					movwf		BD1OBC
					banksel	BD1OST
					movf		BD1OST, W, BANKED		; get EP2 IN status register
					andlw		0x40						; extract the DATA01 bit
					xorlw		0x40						; toggle the DATA01 bit
					iorlw		0x88					; set UOWN and DTS bit
					movwf		BD1OST, BANKED		; prepare reception of the next packet
					call		SendLongBuffer ; start transmission in EP2
					break
				case EP2
					break
			ends
			return

SendDescriptorPacket
			banksel		USB_bytes_left
			ifl USB_bytes_left, LT, 8
				movlw		NO_REQUEST
				movwf		USB_dev_req, BANKED		; sending a short packet, so clear device request
				movf		USB_bytes_left, W, BANKED
			otherwise
				movlw		0x08
			endi
			subwf		USB_bytes_left, F, BANKED
			movwf		USB_packet_length, BANKED
			banksel		BD0IBC
			movwf		BD0IBC, BANKED			; set EP0 IN byte count with packet size
			movf		BD0IAH, W, BANKED		; put EP0 IN buffer pointer...
			movwf		FSR0H, ACCESS
			movf		BD0IAL, W, BANKED
			movwf		FSR0L, ACCESS			; ...into FSR0
			banksel		USB_loop_index
			forlf USB_loop_index, 1, USB_packet_length
				call		Descriptor			; get next byte of descriptor being sent
				movwf		POSTINC0			; copy to EP0 IN buffer, and increment FSR0
				incf		USB_desc_ptr, F, BANKED	; increment the descriptor pointer
			next USB_loop_index
			banksel		BD0IST
			movlw		0x40
			xorwf		BD0IST, W, BANKED		; toggle the DATA01 bit
			andlw		0x40					; clear the PIDs bits
			iorlw		0x88					; set UOWN and DTS bits
			movwf		BD0IST, BANKED
			return

SendKeyBuffer
			banksel		BD1IAH
			movf		BD1IAH, W, BANKED		; put EP1 IN buffer pointer...
			movwf		FSR0H, ACCESS
			movf		BD1IAL, W, BANKED
			movwf		FSR0L, ACCESS			; ...into FSR0
			banksel		Key_buffer
			movf		Key_buffer, W, BANKED	; copy modifier byte...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		Key_buffer+1, W, BANKED	; copy reserved byte...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		Key_buffer+2, W, BANKED	; copy keycode 1...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		Key_buffer+3, W, BANKED	; copy keycode 2...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		Key_buffer+4, W, BANKED	; copy keycode 3...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		Key_buffer+5, W, BANKED	; copy keycode 4...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		Key_buffer+6, W, BANKED	; copy keycode 5...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		Key_buffer+7, W, BANKED	; copy keycode 6...
			movwf		INDF0					; ...to EP1 IN buffer
			banksel		BD1IBC
			movlw		0x08
			movwf		BD1IBC, BANKED			; set EP1 IN buffer byte count to 8
			movf		BD1IST, W, BANKED		; get EP1 IN status register
			andlw		0x40					; extract the DATA01 bit
			xorlw		0x40					; toggle the DATA01 bit
			iorlw		0x88					; set UOWN and DTS bits
			movwf		BD1IST, BANKED			; send packet
			return
; Copy buffer now changed
; source in BD10A(HL), destination in FSR1(HL)
; in W - value added to the buffer
CopyBuffer
			movff		BD1OAH, FSR0H		; put EP1 IN buffer pointer...
			movff		BD1OAL, FSR0L		; ...into FSR0
			movff		BD1OBC,BD2IBC
			movff		WREG, COUNTER
loop_11
			movf		POSTINC0, W, ACCESS
			banksel	COUNTER
			addwf		COUNTER, W, BANKED
			movwf    POSTINC1,ACCESS
			banksel  BD2IBC
			decfsz   BD2IBC,F
			goto 		loop_11
			return

; The routines below are sending the long response
SendLongBuffer:
		; Check if there is more than one buffer to send
			banksel USB_LB_running
			ifl USB_LB_running, EQ, 0
				return
			endi
			banksel USB_LB_bytes_to_send
			ifl USB_LB_bytes_to_send+1, EQ, 0
				andif USB_LB_bytes_to_send, LT, 0x40
					banksel 	USB_LB_running
					clrf		USB_LB_running, BANKED	; clear device request
					banksel	USB_LB_bytes_to_send
					movf		USB_LB_bytes_to_send, W, BANKED
			otherwise
					movlw		0x40
			endi
			banksel  BD2IBC
			movwf		BD2IBC, BANKED
		; Put the buffer location into the transfer address
			movff		USB_LB_ptr_to_send, BD2IAL
			movff		USB_LB_ptr_to_send+1, BD2IAH
		; Now update the number of bytes to send
			banksel	USB_LB_bytes_to_send
			subwf		USB_LB_bytes_to_send, F
			movlw		0
			subwfb	USB_LB_bytes_to_send+1, F
		; Now update the pointer location
			movff		BD2IBC, WREG
			banksel	USB_LB_ptr_to_send
			addwf		USB_LB_ptr_to_send,F
			movlw		0
			addwfc	USB_LB_ptr_to_send+1,F
		; Initiate the transfer
			banksel	BD2IST
			movf		BD2IST, W, BANKED		; get EP2 IN status register
			andlw		0x40					; extract the DATA01 bit
			xorlw		0x40					; toggle the DATA01 bit
			iorlw		0x88					; set UOWN and DTS bits
			movwf		BD2IST, BANKED			; send packet
			return

write_samples:
			call buf_read ; C - error
			btfsc STATUS,C,ACCESS
			return
			call sd_write_byte ; C - error
			return

APPLICATION	code
Main
		banksel sampling_flag
		clrf sampling_flag
		banksel		COUNTER
 		for COUNTER, 0x01, 0x17		; do nothing for 16-17 us
		next COUNTER
		clrf	PORTA, ACCESS
		movlw	0x0F
		movwf	ADCON1, ACCESS		; set up PORTA to be digital I/Os
		movlw	0xff
		movwf	TRISA, ACCESS		; make port A all logic inputs for now

 		movlw	b'01000000'			; PORT D for MSPI (bits 7-4="0100") and for LCD (bits 3-0="0000")
 		movwf	TRISD, ACCESS
		movlw b'11111111'
		;movwf TRISB,ACCESS
		movlw b'00000111'
		movwf TRISE,ACCESS
		movf	PORTB, W, ACCESS
		iorlw	b'00010000'			; make RB4 an input (SW2)
		movwf	PORTB, ACCESS
; Settings for the TRISC
; EUART requires b7=b6=1, LCD requires b2=b1=b0=0
		movlw b'11111000'; 
 		movwf	TRISC, ACCESS
		clrf PORTC,ACCESS
;;; lp1		btg		PORTD, 1, ACCESS
;;;  		for COUNTER, 0x01, 0x17		; do nothing for 16-17 us
;;; 		next COUNTER
;;;  		goto	lp1
		
;        movlw		TIMER0H_VAL
;		movwf		TMR0H, ACCESS
;        movlw		TIMER0L_VAL
;		movwf		TMR0L, ACCESS
;		movlw		0x97
;		movwf		T0CON, ACCESS			; set prescaler for Timer0 for 1:256 scaling
											;	(Timer0 will go off every ~10 ms )
		movlw 	0xfc
		call		DEL_N_MS
		call		LCD_INIT
		call		buf_init
		banksel	COUNTER
		movlw		0xf0
		movwf		COUNTER,BANKED
main_sd_init_loop1:
		call		sd_init
		bnc main_sd_init_ok
		movlw 	0x2
		call		DEL_N_MS
		decfsz	COUNTER,BANKED
		goto main_sd_init_loop1
		goto main_sd_init_error
main_sd_init_ok:
		call sd_calc_max_block
; Now we find the first free block
		call sd_find_free
main_sd_init_error:
		btfsc		PORTE,2,ACCESS
		  call		InitUSB					; initialize the USB registers and serial interface engine
		call		ser_init
		call		adc_init
		; adc_init has set up the timer1, now we connect additionally our timer routine to it
		;bcf PIR1,TMR1IF,ACCESS
		;bsf PIE1,TMR1IE,ACCESS
		;bsf IPR1,TMR1IP,ACCESS
;Interrupts!
;		movlw '@'
;		movff WREG,0x300
;		movlw '#'
;		movff WREG,0x301
		banksel int_flag
		clrf int_flag, BANKED
		bcf RCON,IPEN,ACCESS
		clrf INTCON2, ACCESS
		;bsf  INTCON2, RBPU, ACCESS
		bsf INTCON,INT0IE, ACCESS
		bsf INTCON,GIE, ACCESS ; GIEH enabled
		bsf INTCON,PEIE, ACCESS ; GIEH enabled
		btfss PORTE,2,ACCESS
		goto record_data
		repeat
			call		ServiceUSB			; service USB requests...
; 			banksel		PORTD
; 			btg			PORTD, 1, ACCESS
			banksel		USB_USWSTAT
		until USB_USWSTAT, EQ, CONFIG_STATE	; ...until the host configures the peripheral

;;; 		bsf		PORTD, 1, ACCESS	; light up D2 DIAGNOSTIC

		banksel		COUNTER
		lfsr  2,lcd_row_0
		movlw 'U'
		movwf	POSTINC2,ACCESS
		movlw 'S'
		movwf	POSTINC2,ACCESS
		movlw 'B'
		movwf	POSTINC2,ACCESS
		movlw ' '
		movwf	POSTINC2,ACCESS
		movlw ' '
		movwf	POSTINC2,ACCESS
		movlw ' '
		movwf	POSTINC2,ACCESS
		movwf	POSTINC2,ACCESS
		movwf	POSTINC2,ACCESS
		movlw ' '
		repeat
;			repeat
				call		ServiceUSB			; service USB requests...
				banksel sampling_flag
				tstfsz  sampling_flag
				call write_samples
;			untilset INTCON, T0IF, ACCESS		; ...until Timer0 goes off
;			bcf			INTCON, T0IF, ACCESS	; clear Timer0 interrupt flag
;           movlw		TIMER0H_VAL
;		movwf		TMR0H, ACCESS
;        movlw		TIMER0L_VAL
;	movwf		TMR0L, ACCESS
;			banksel		BD1IST
;			ifclr BD1IST, UOWN, BANKED			; check to see if the PIC owns the EP1 IN buffer
;				movlw		high (Key_buffer+2)
;				movwf		FSR0H, ACCESS
;				movlw		low (Key_buffer+2)
;				movwf		FSR0L, ACCESS		; set FSR0 to point to start of keycodes in Key_buffer
;				ifset PORTB, 4, ACCESS			; see if SW2
;					movlw		0x39			; USB keycode for CAPS LOCK
;					movwf		POSTINC0		; put CAPS LOCK keycode into Key_buffer
;				endi
;				call		GetNextKeycode		; get the next keycode and...
;				movwf		POSTINC0			; ...put it into Key_buffer
;				clrf		INDF0
;				incf		COUNTER, F, BANKED	; increment COUNTER...
;				movlw		0x07
;				andwf		COUNTER, F, BANKED	; ...modulo 8
;				call		SendKeyBuffer
;			endi
		forever
		
record_data:
			; Wait for button 1
			lfsr  2,lcd_row_0
			movlw 'R'
			movwf	POSTINC2,ACCESS
			movlw 'e'
			movwf	POSTINC2,ACCESS
			movlw 'a'
			movwf	POSTINC2,ACCESS
			movlw 'd'
			movwf	POSTINC2,ACCESS
			movlw 'y'
			movwf	POSTINC2,ACCESS
			movlw ' '
			movwf	POSTINC2,ACCESS
			movwf	POSTINC2,ACCESS
			movwf	POSTINC2,ACCESS
wait_but_1:
			btfsc PORTB,5,ACCESS
			goto wait_but_1
			lfsr  2,lcd_row_0
			movlw 'R'
			movwf	POSTINC2,ACCESS
			movlw 'e'
			movwf	POSTINC2,ACCESS
			movlw 'c'
			movwf	POSTINC2,ACCESS
			movlw 'o'
			movwf	POSTINC2,ACCESS
			movlw 'r'
			movwf	POSTINC2,ACCESS
			movlw 'd'
			movwf	POSTINC2,ACCESS
			movlw ' '
			movwf	POSTINC2,ACCESS
			movwf	POSTINC2,ACCESS
			movff sd_test_block+3,sd_address+3
			movff sd_test_block+2,sd_address+2
			movff sd_test_block+1,sd_address+1
			movff sd_test_block+0,sd_address+0
			call sd_write_init
			movlw B'11000000'
			call buf_write_user
			movlw B'11000000'
			call buf_write_user
			banksel sampling_flag
			setf sampling_flag,BANKED
record_data1:
			call write_samples
			btfsc	PORTB,6,ACCESS
			goto record_data1
			banksel sampling_flag
			clrf sampling_flag,BANKED
			movlw B'11000001'
			call buf_write_user
			movlw B'11000001'
			call buf_write_user
			; transfer all bytes remaing in the buffer
record_data2:
			call write_samples
			btfss STATUS,C,ACCESS
			goto record_data2
			call	sd_write_stop
			call sd_find_free
			goto record_data

GetNextKeycode
			movlw		upper KeycodeTable
			movwf		TBLPTRU, ACCESS
			movlw		high KeycodeTable
			movwf		TBLPTRH, ACCESS
			movlw		low KeycodeTable
			banksel		COUNTER
			addwf		COUNTER, W, BANKED
			ifset STATUS, C, ACCESS
				incf		TBLPTRH, F, ACCESS
				ifset STATUS, Z, ACCESS
					incf		TBLPTRU, F, ACCESS
				endi
			endi
			movwf		TBLPTRL, ACCESS
			tblrd*
			movf		TABLAT, W, ACCESS
			return

KeycodeTable
			db			0x09, 0x12	; USB keycode for 'f', USB keycode for 'o'
			db			0x00, 0x12	; USB keycode for indicating no event, USB keycode for 'o'
			db			0x05, 0x04	; USB keycode for 'b', USB keycode for 'a'
			db			0x15, 0x2C	; USB keycode for 'r', USB keycode for ' '
HiIrq:	
			pop ; According to the errata 80220g.pdf
			banksel int_flag
			setf int_flag, BANKED
			movff FSR2L,fsr2_save
			movff FSR2H,fsr2_save+1
			movff buf_wr_ptr,FSR2L
			movff buf_wr_ptr+1,FSR2H
			call ser_rx_irq
			call int0_irq
			call adc_irq
			movff FSR2L,buf_wr_ptr
			movff FSR2H,buf_wr_ptr+1
			;call  LCD_INT ; uses the IND2 registers
			movff fsr2_save,FSR2L
			movff fsr2_save+1,FSR2H
			retfie FAST

int0_irq:
			btfss INTCON,INT0IF,ACCESS
			return
			bcf   INTCON,INT0IE,ACCESS
			return
			end
