; The main problem with the LCD is that it requires quite a lot of time to send a byte
; Unfortunately, the LCD_E cycle must be not faster than 1us, with high phase not 
; shorter than 450 ns. It forces us to use the active waiting, when sending data 
; to the LCD. In turn it requires us not to update the LCD when recording the voice :-(
;
; However, what if I could assume controlling LCD with much lower frequency?
; If I had to refresh LCD with 32 characters 10 times per second, then I would need to write
; only ca 1000 nybbles per second.
; If I have a 8kHz interrupts used to sample the voice, then I can use the same interrupts 
; to control the LCD
	list      p=18f4550            ; list directive to define processor
#include <p18f4550.inc>
#include lcd/lcd.inc
#include supp.inc
LCD_DATA udata
lcd_row_0 res D'16'
lcd_row_1 res D'16'
lcd_cmd_phase res 1
lcd_data res 1
lcd_row res 1
lcd_col res 1
LCD_COLS equ D'16'
	global lcd_row, lcd_col,lcd_data,lcd_cmd_phase
	global lcd_row_0
	global lcd_row_1
LCD_CODE CODE
; The new implementation will use the 8kHz interrupt driven state machine
; However the problem is: how to create the routine, which will 
; be run after each interrupt, and suspended till the next interrupt (or even
; for a longer time - e.g. 40 ms!)

lcd_sched:
	global lcd_sched
	; The routine of the LCD driver interpreter
	; called 8000 times per second
	;
	; Check the current state of the interpreter
	;  e.g. we can be in the state "waiting for delay"
	; Read the next instruction and execute it...
	; The commands may be read with the
	; TBLRD*+ instruction 
	; In fact we need only a few instructions:
	; Send byte with RS=0, Send byte with RS=1
	; Wait for a while
	; Each instruction will have a few phases
	; 1) Send lower nibble, E->0
	; 2) Send lower nibble, E->1
	; 3) Send higher nibble, E->0
	; 4) Send higher nibble, E->1
	return
LCD_CODE0 CODE

LCD_INIT2:
  global LCD_INIT2
  ; 1.  Wait more than 15 msecs after power is applied.
;   lcd_del_ms 15
;	lcd_write4_rs0 0x03
;	lcd_del_ms 5
;	lcd_write4_rs0 0x03
;	lcd_del_us D'160'
;	lcd_write4_rs0 0x03
;	lcd_del_us D'160'
   ;* Write 0x02 to the LCD to Enable Four Bit Mode
;  lcd_write4_rs0 0x02
;	lcd_del_ms D'41'
   ; Starting from this point we have to perform 2-nibble transfers
   ; Set the interface length to 0x28 (4-bit interface, 2 lines, 5x7 font)
;	lcd_write8_rs0 0x28
;	lcd_del_ms D'41'
   ; Write 0x01/0x00 to turn off the Display
;	lcd_write8_rs0 0x02
;	lcd_del_ms D'41'
   ; Write 0x01/0x00 to turn off the Display
;   lcd_write8_rs0 0x01
;	lcd_del_ms D'41'
   ; Write 0x00/0x01 to Clear the Display
;   lcd_write8_rs0 0x06
;	lcd_del_ms D'41'
;   return



LCD_WRITE4_RS0:
   global LCD_WRITE4_RS0
   ; Shift the data
   andlw 0xf
   mullw (1 << LCD_DSHIFT)
   ; Shifted data are located in the PRODL
   ; Zero the data bits in the port
   movlw 0xff - ( 0xf << LCD_DSHIFT )
   andwf LCD_DPORT,W
   iorwf PRODL,W
   movwf LCD_DPORT
   bcf   LCD_RS_PORT, LCD_RS_PIN
   bcf   LCD_RW_PORT, LCD_RW_PIN
   bsf   LCD_E_PORT, LCD_E_PIN
   movlw 1
   call  DEL_N_MS
   bcf   LCD_E_PORT, LCD_E_PIN
   movlw 1
   call  DEL_N_MS
   return
LCD_WRITE4_RS1:
   global LCD_WRITE4_RS1
   ; Shift the data
   andlw 0xf
   mullw (1 << LCD_DSHIFT)
   ; Shifted data are located in the PRODL
   ; Zero the data bits in the port
   movlw 0xff - ( 0xf << LCD_DSHIFT )
   andwf LCD_DPORT,W
   iorwf PRODL,W
   movwf LCD_DPORT
   bsf   LCD_RS_PORT, LCD_RS_PIN
   bcf   LCD_RW_PORT, LCD_RW_PIN
   bsf   LCD_E_PORT, LCD_E_PIN
   movlw 1
   call  DEL_N_MS
   bcf   LCD_E_PORT, LCD_E_PIN
   movlw 1
   call  DEL_N_MS
   return
LCD_ACC_DATA UDATA_ACS
LCD_TMP1 RES 1
LCD_CODE1 CODE
LCD_WRITE8_RS0:
   global LCD_WRITE8_RS0
   movwf LCD_TMP1
   andlw 0xf0
   mullw (1<<4)
   movf PRODH,W
   call LCD_WRITE4_RS0
   movf LCD_TMP1,W
   call LCD_WRITE4_RS0
   return
LCD_WRITE8_RS1:
   global LCD_WRITE8_RS1
   movwf LCD_TMP1
   andlw 0xf0
   mullw (1<<4)
   movf PRODH,W
   call LCD_WRITE4_RS1
   movf LCD_TMP1,W
   call LCD_WRITE4_RS1
   return

LCD_INIT:
  global LCD_INIT
  ; First - fill the LCD buffer
  ;
	banksel lcd_cmd_phase
	clrf lcd_cmd_phase, BANKED
	banksel lcd_col
	clrf lcd_col, BANKED
	banksel lcd_row
	clrf lcd_row, BANKED
   LFSR 2,lcd_row_0
	movlw 0x20
	movff WREG,lcd_data
	movlw 0x10
lcd_init_loop_1:
	movff lcd_data,POSTINC2
	decfsz WREG,ACCESS
	goto lcd_init_loop_1
   LFSR 2,lcd_row_1
	movlw 0x10
lcd_init_loop_2:
	movff lcd_data,POSTINC2
	decfsz WREG,ACCESS
	goto lcd_init_loop_2
	LFSR 2,lcd_row_0
	movlw 'G'
	movwf POSTINC2,ACCESS
	movlw 'P'
	movwf POSTINC2,ACCESS
	movlw 'S'
	movwf POSTINC2,ACCESS
	movlw ' '
	movwf POSTINC2,ACCESS
	movlw 'U'
	movwf POSTINC2,ACCESS
	movlw 'S'
	movwf POSTINC2,ACCESS
	movlw 'B'
	movwf POSTINC2,ACCESS
   LFSR 2,lcd_row_1
	movlw 'L'
	movwf POSTINC2,ACCESS
	movlw 'o'
	movwf POSTINC2,ACCESS
	movlw 'g'
	movwf POSTINC2,ACCESS
	movlw 'g'
	movwf POSTINC2,ACCESS
	movlw 'e'
	movwf POSTINC2,ACCESS
	movlw 'r'
	movwf POSTINC2,ACCESS
	movlw ' '
	movwf POSTINC2,ACCESS
; 1.  Wait more than 15 msecs after power is applied.
   movlw 15
   call DEL_N_MS
  ; 2. Write 0x03 to LCD and wait 5 msecs for the instruction to complete
   movlw 0x03
   call LCD_WRITE4_RS0
   movlw 5
   call DEL_N_MS
   ; 3. Write 0x03 to LCD and wait 160 usecs for instruction to complete
   movlw 0x03
   call LCD_WRITE4_RS0
   movlw D'160'
   call DEL_N_US
   ; 4. Write 0x03 AGAIN to LCD and wait 160 usecs (or poll the Busy Flag)
   movlw 0x03
   call LCD_WRITE4_RS0
   movlw D'160'
   call DEL_N_US
   ;5. Set the Operating Characteristics of the LCD
   ;* Write 0x02 to the LCD to Enable Four Bit Mode
   movlw 0x02
   call LCD_WRITE4_RS0
   movlw D'41'
   call DEL_N_MS
   ; Starting from this point we have to perform 2-nibble transfers
   ; Set the interface length to 0x28 (4-bit interface, 2 lines, 5x7 font)
   movlw 0x28
   call LCD_WRITE8_RS0
   movlw D'41'
   call DEL_N_MS
   ; Write 0x01/0x00 to turn off the Display
   movlw 0x02
   call LCD_WRITE8_RS0
   movlw D'41'
   call DEL_N_MS
   ; Write 0x0f/0x00 to turn on the Display
   movlw 0x0c
   call LCD_WRITE8_RS0
   movlw D'41'
   call DEL_N_MS
   ; Write 0x00/0x01 to Clear the Display
   movlw 0x06
   call LCD_WRITE8_RS0
   movlw D'41'
   call DEL_N_MS
	;call LCD_TEST
   return

LCD_TEST:
   global LCD_TEST
   movlw D'250'
   call DEL_N_MS
   movlw D'250'
   call DEL_N_MS
   movlw D'250'
   call DEL_N_MS
   movlw 0x41
   call LCD_WRITE8_RS1
   movlw 41
   call DEL_N_MS
   movlw 0x42
   call LCD_WRITE8_RS1
   movlw 41
   call DEL_N_MS
   movlw 0x43
   call LCD_WRITE8_RS1
   movlw 41
   call DEL_N_MS
   movlw 0x44
   call LCD_WRITE8_RS1
   movlw 41
   call DEL_N_MS
	goto $
   return

; The procedure below is the interrupt driven version of the LCD refresh
; After each interrupt we generate half of the write cycle.
; Depending on the counter setting
; Constants used: LCD_ROW_LENGTH - length of the single row
; LCD_ROWS_NUM - number of the LCD rows
; Single byte is transferred in 4 cycles. Each nibble is transferred in 2 cycles
; In the first cycle the data, the RS and RW lines are set, and then E is set to high
; In the second cycle the E is set to low.
; So the bit0 of the cycle counter may be used to set the E state,
; the bit1 of the cycle counter may be used to set the nybble number
; the rest may be used to count the data number, however we need to
; set the DRAM address on the end of the first row, and on the end of the second row...
; When the number of character is equal to the length of row, then we send the command
; to move the cursor to the begining of the next row
LCD_INT:
	global LCD_INT
   ;In the LCD_INT_CNT we hold the following information:
   ; bit 7 - row number
   ; bits 6-2 number of the character in the row (0-31 max, however one position must be reserved for
   ; end of row)
   ; bits 1 and 0 - phase number: bit 0 - inverted state of the E line, bit 1 - nybble number
 
   ;In the MSB we keep the row number
   ; First check the command phase
	;btfss PIR1,TMR1IF,ACCESS
	;return
	;bcf PIR1,TMR1IF,ACCESS
	banksel lcd_cmd_phase
	incf lcd_cmd_phase,W,BANKED
	dcfsnz WREG,BANKED
	goto lcd_phase_0
	dcfsnz WREG,BANKED
	goto lcd_phase_1
	dcfsnz WREG,BANKED
	goto lcd_phase_2
	goto lcd_phase_3
lcd_phase_0:
	banksel lcd_col
	; Check the character number
	movlw LCD_COLS
	cpfslt lcd_col,BANKED
	goto lcd_phase_0_next_row
	; Check the row number (0 or 1)
	lfsr	2,lcd_row_0
	banksel lcd_row
	tstfsz lcd_row,BANKED
	lfsr	2,lcd_row_1
	; Now we have the proper row buffer in the INDF2
	banksel lcd_col
	movf lcd_col,W,BANKED
	incf lcd_col,F,BANKED
	movff PLUSW2,lcd_data
	; Set the line RS to 1
	bsf   LCD_RS_PORT, LCD_RS_PIN,ACCESS
	goto lcd_phase_0_send_data
lcd_phase_0_next_row:
	; Check the row number (0 or 1)
	movlw b'11000000' ; Go to position 0x40 command
	banksel lcd_row
	tstfsz lcd_row
	movlw b'10000000' ; Go to position 0x0 command
	movff WREG,lcd_data
	; Adjust the row number
	btg lcd_row,0,BANKED
	; Set the character number to 0
	banksel lcd_col
	clrf lcd_col,BANKED
	; Set the line RS to 0
 	bcf   LCD_RS_PORT, LCD_RS_PIN,ACCESS
lcd_phase_0_send_data:
	movff lcd_data, WREG
   bcf   LCD_RW_PORT, LCD_RW_PIN,ACCESS
	; Send the first nibble
   mullw (1<<4)
   movf PRODH,W
   ; Shift the data (they are still in the WREG!)
   mullw (1 << LCD_DSHIFT)
   ; Shifted data are located in the PRODL
   ; Zero the data bits in the port
   movlw 0xff - ( 0xf << LCD_DSHIFT )
   andwf LCD_DPORT,W,ACCESS
   iorwf PRODL,W,ACCESS
   movwf LCD_DPORT,ACCESS
	; increase the phase
	banksel lcd_cmd_phase
	incf lcd_cmd_phase,BANKED
   bsf   LCD_E_PORT, LCD_E_PIN,ACCESS
	return
lcd_phase_1:
	banksel lcd_cmd_phase
   bcf   LCD_E_PORT, LCD_E_PIN,ACCESS
	incf lcd_cmd_phase,BANKED
	return
lcd_phase_2:
	movff lcd_data,WREG
	andlw 0xf
	mullw (1 << LCD_DSHIFT)
   ; Shifted data are located in the PRODL
   ; Zero the data bits in the port
   movlw 0xff - ( 0xf << LCD_DSHIFT )
   andwf LCD_DPORT,W
   iorwf PRODL,W
   movwf LCD_DPORT,ACCESS
	; increase the phase
	banksel lcd_cmd_phase
	incf lcd_cmd_phase,BANKED
   bsf   LCD_E_PORT, LCD_E_PIN,ACCESS
	return
lcd_phase_3:
	banksel lcd_cmd_phase
	clrf lcd_cmd_phase,BANKED
   bcf   LCD_E_PORT, LCD_E_PIN,ACCESS
	return
   ;
	end