	list      p=18f4550            ; list directive to define processor
	#include <p18f4550.inc>        ; processor specific variable definitions
	extern buf_init, buf_read, buf_write
	extern LCD_INT

bank0 udata
adc_ptr res 2
adc_val res 2
last_adc_ptr res 1
button_count res 1
fsr2_sav res 2

	global adc_ptr, last_adc_ptr

ADC_CODE CODE
adc_init:
	global adc_init
	banksel button_count
	clrf button_count, BANKED
	;Initialize the pointer to the buffer
	banksel adc_ptr
	movlw 0x03
	movwf adc_ptr+1
	movlw 0x00
	movwf adc_ptr
	banksel last_adc_ptr
	movwf last_adc_ptr
	;Program the CCP2 timer for 8kHz sampling rate
	;Set the input channel 0 as a digital input
	movlw 0x03
	iorwf TRISA,F,ACCESS
	; set the ADCON1 b3:0 - PCFG=0xe (only AN0 is analog), Vref comes from AVDD and AVSS
	movlw 0x0d 
	movwf ADCON1,ACCESS
	;
	bcf PIR1,ADIF,ACCESS ; Clear A/D interrupt flag bit
	bsf PIE1,ADIE,ACCESS ; Enable A/D interrupts
  	bsf IPR1,ADIP,ACCESS ; High Priority
	; Set the channel 
	movlw 0x05  ; b0=1 (ADC on), b5-b2=1(4) (channel 0)
  	movwf ADCON0,ACCESS
	movlw 0x06  ; Left Justified, TAD=0, clock=Fosc/64
	movwf ADCON2,ACCESS
	; Program the timer
   movlw high(d'12000000'/d'8000')
	movwf CCPR2H,ACCESS
   movlw low(d'12000000'/d'8000')
	movwf CCPR2L,ACCESS
  	; Special event trigger from CCP2 will start the conversion
  	movlw 0x0b
	movwf CCP2CON,ACCESS
   ; Start the timer
   clrf T3CON,ACCESS ; Necessary to connect CCP2 to timer1
   movlw 0x04 ; 
	movwf T1CON,ACCESS
   clrf TMR1L,ACCESS
   clrf TMR1H,ACCESS
	movlw 0x01 ; Switch the timer 1 on
   movwf T1CON,ACCESS
	return

adc_irq:
	global adc_irq
	; Check if the ADC interrupt is pending
	btfss PIR1,ADIF,ACCESS
	return
	; We receive the data from the ADC
	movff ADRESH,adc_val+1
	movff ADRESL,adc_val
	bcf PIR1,ADIF,ACCESS
	; Check if the button is pressed
	banksel button_count
	btfsc PORTB,0,ACCESS
	goto adc_irq_no_button
	; button is pressed, if button_count < 0xff, increase it
	comf button_count,W,BANKED
	btfss STATUS,Z,ACCESS
	incf button_count,F,BANKED
   ;if voice recording is on, continue
   btfsc button_count,7,BANKED
   goto adc_irq_voice_test
	; if voice recording is off and the result is bigger, than 0x10, set button_count to 0xff
   movlw 0x10
   cpfsgt button_count,BANKED
   goto adc_irq_voice_test
   setf button_count,BANKED
   ; Send the marker of the message begining
   movlw B'10000000'
   call buf_write
   movlw B'10000000'
   call buf_write
   goto adc_irq_voice_test
adc_irq_no_button:	
	; button is released, if button_count > 0, decrease it
	tstfsz button_count,BANKED
	decf button_count,F,BANKED
	; If voice recording is off, continue
	btfss button_count,7,BANKED
	goto adc_irq_voice_test
	; if result is smaller than 0xf0, set button_count to 0
	movlw 0xf0
	cpfslt button_count,BANKED
	goto adc_irq_voice_test
	clrf button_count,BANKED
	; Send the marker of the end of the message
   movlw B'10000001'
   call buf_write
   movlw B'10000001'
   call buf_write
adc_irq_voice_test	
	banksel button_count
	btfss button_count,7,BANKED
	goto adc_irq_no_voice_rec
	banksel adc_val
	rrcf adc_val+1,F,BANKED
	rrcf adc_val,F,BANKED
	rrcf adc_val+1,F,BANKED
	rrcf adc_val,F,BANKED
	rrcf adc_val,F,BANKED
	rrcf adc_val,F,BANKED
	bsf  adc_val+1,7,BANKED
	bsf  adc_val+1,6,BANKED
	bsf  adc_val,7,BANKED
	bcf  adc_val,6,BANKED
	movff adc_val+1,WREG
	call buf_write
	movff adc_val,WREG
	call buf_write
adc_irq_no_voice_rec:
	movff FSR2L,fsr2_sav
	movff FSR2H,fsr2_sav+1
	call LCD_INT
	movff fsr2_sav,FSR2L
	movff fsr2_sav+1,FSR2H
	return
END